# Link to the current official Wwise doc page:
# https://www.audiokinetic.com/en/public-library/2024.1.7_8863/?source=SDK&id=_ak_speaker_config_8h_source.html

speaker_config_options = {
    "AK_SPEAKER_SETUP_MONO": 1,
    "AK_SPEAKER_SETUP_0POINT1": 1,
    "AK_SPEAKER_SETUP_1POINT1": 2,
    "AK_SPEAKER_SETUP_STEREO": 2,
    "AK_SPEAKER_SETUP_2POINT1": 3,
    "AK_SPEAKER_SETUP_3STEREO": 3,
    "AK_SPEAKER_SETUP_3POINT1": 4,
    "AK_SPEAKER_SETUP_4": 4,
    "AK_SPEAKER_SETUP_4POINT1": 5,
    "AK_SPEAKER_SETUP_5": 5,
    "AK_SPEAKER_SETUP_5POINT1": 6,
    "AK_SPEAKER_SETUP_6": 6,
    "AK_SPEAKER_SETUP_6POINT1": 7,
    "AK_SPEAKER_SETUP_7": 7,
    "AK_SPEAKER_SETUP_7POINT1": 8,
    "AK_SPEAKER_SETUP_SURROUND": 3,
    "AK_SPEAKER_SETUP_HEIGHT_2": 2,
    "AK_SPEAKER_SETUP_HEIGHT_4": 4,
    "AK_SPEAKER_SETUP_HEIGHT_5": 5,
    "AK_SPEAKER_SETUP_HEIGHT_ALL": 6,
    "AK_SPEAKER_SETUP_HEIGHT_4_TOP": 5,
    "AK_SPEAKER_SETUP_HEIGHT_5_TOP": 6,
    "AK_SPEAKER_SETUP_AURO_222": 6,
    "AK_SPEAKER_SETUP_AURO_8": 8,
    "AK_SPEAKER_SETUP_AURO_9": 9,
    "AK_SPEAKER_SETUP_AURO_9POINT1": 10,
    "AK_SPEAKER_SETUP_AURO_10": 10,
    "AK_SPEAKER_SETUP_AURO_10POINT1": 11,
    "AK_SPEAKER_SETUP_AURO_11": 11,
    "AK_SPEAKER_SETUP_AURO_11POINT1": 12,
    "AK_SPEAKER_SETUP_AURO_11_740": 11,
    "AK_SPEAKER_SETUP_AURO_11POINT1_740": 12,
    "AK_SPEAKER_SETUP_AURO_13_751": 13,
    "AK_SPEAKER_SETUP_AURO_13POINT1_751": 14,
    "AK_SPEAKER_SETUP_DOLBY_5_0_2": 7,
    "AK_SPEAKER_SETUP_DOLBY_5_1_2": 8,
    "AK_SPEAKER_SETUP_DOLBY_5_0_4": 9,
    "AK_SPEAKER_SETUP_DOLBY_5_1_4": 10,
    "AK_SPEAKER_SETUP_DOLBY_6_0_2": 8,
    "AK_SPEAKER_SETUP_DOLBY_6_1_2": 9,
    "AK_SPEAKER_SETUP_DOLBY_6_0_4": 10,
    "AK_SPEAKER_SETUP_DOLBY_6_1_4": 11,
    "AK_SPEAKER_SETUP_DOLBY_7_0_2": 9,
    "AK_SPEAKER_SETUP_DOLBY_7_1_2": 10,
    "AK_SPEAKER_SETUP_DOLBY_7_0_4": 11,
    "AK_SPEAKER_SETUP_DOLBY_7_1_4": 12,
    "AK_SPEAKER_SETUP_0_1": 1,
    "AK_SPEAKER_SETUP_1_0_CENTER": 1,
    "AK_SPEAKER_SETUP_1_1_CENTER": 2,
    "AK_SPEAKER_SETUP_2_0": 2,
    "AK_SPEAKER_SETUP_2_1": 3,
    "AK_SPEAKER_SETUP_3_0": 3,
    "AK_SPEAKER_SETUP_3_1": 4,
    "AK_SPEAKER_SETUP_FRONT": 3, 
    "AK_SPEAKER_SETUP_4_0": 4,
    "AK_SPEAKER_SETUP_4_1": 5,
    "AK_SPEAKER_SETUP_5_0": 5,
    "AK_SPEAKER_SETUP_5_1": 6,
    "AK_SPEAKER_SETUP_6_0": 6,
    "AK_SPEAKER_SETUP_6_1": 7,
    "AK_SPEAKER_SETUP_7_0": 7,
    "AK_SPEAKER_SETUP_7_1": 8
}